using System.Collections.Generic
using System

public abstract class Rabit[T] where T : Rabit[T]
  public abstract CrossOver (other : T) : T
  public abstract MutateImpl () : void
  public abstract JudgeImpl () : double

  mutable dirty = true
  mutable score : double

  public Score : double
    get
      when (dirty) score = JudgeImpl ()
      dirty = false
      score

  public Mutate () : void
    dirty = true
    MutateImpl ()

  class Comparer : IComparer [T]
    public Compare (t1 : T, t2 : T) : int
      t2.Score.CompareTo (t1.Score)

  static public Comparator : IComparer [T] = Comparer ()

public class Population[T] where T : Rabit[T]
  mutatation_factor = 0.6
  survival_factor = 0.1
  max_population = 2000
  round_count = 500

  rabits : array [T] = array (max_population)

  rand : Random = Random ()

  mutable iter = 0

  Born (max : int) : T
    def p1 = rabits [rand.Next (max)]
    def p2 = rabits [rand.Next (max)]
    p1.CrossOver (p2)

  OneRound () : void
    iter++
    Array.Sort (rabits, Rabit[T].Comparator)
    Console.WriteLine ($ "round $iter, best at $(rabits[0].Score)")
    mutable last_rabit = (max_population * survival_factor) :> int
    while (last_rabit < max_population)
      rabits [last_rabit] = Born (last_rabit)
      when (rand.NextDouble () < mutatation_factor)
        rabits [last_rabit].Mutate ()
      last_rabit++

  FillIt (count : int) : void
    mutable ptr = count
    while (ptr < max_population)
      rabits [ptr] = Born (count)
      rabits [ptr].Mutate ()
      ptr++

  public Go (first_rabits : list [T]) : void
    mutable i = 0
    foreach (r in first_rabits)
      rabits [i] = r
      i++
    FillIt (i)
    repeat (round_count)
      OneRound ()
    Array.Sort (rabits, Rabit[T].Comparator)
    Console.WriteLine ("---------------------------------------------------------------")
    foreach (k in $[0 .. 3])
      Console.Write ($ "$k. $(rabits [k].Score): $(rabits [k])\n")


