(theory Int_Heaps

 :written_by {Michal Moskal}
 :date {04/01/07}
 
 :sorts (Int Object Heap Array UNKNOWN)

 :funs (
   (select2 Heap Object Object Object) 
   (store2  Heap Object Object Object Heap)
   (int_to_object Int Object)
   (object_to_int Object Int)

   (select Array Object Object) 
   (store  Array Object Object Array)
   (array_to_object Array Object)
   (object_to_array Object Array)

   (0 Int) 
   (1 Int)
   (~ Int Int)             ; unary minus
   (- Int Int Int :assoc) ; binary minus
   (+ Int Int Int :assoc :comm :unit {0}) 
   (* Int Int Int :assoc :comm :unit {1})
 )


 :preds (
         (<= Int Int :refl :trans :antisym)
         (< Int Int :trans :irref)
         (>= Int Int :refl :trans :antisym)
         (> Int Int :trans :irref)
         (subtypes Object Object)
        )

 :definition 
 "This is a theory of functional arrays without extensionality.
  Additionally integers and functions casting between array 
  elements/indices and integers.
 "

 :axioms 
 (
  (forall (?h Heap) (?o Object) (?f Object) (?e Object)
     (= (select2 (store2 ?h ?o ?f ?e) ?o ?f) ?e))

  (forall (?h Heap) (?o Object) (?f Object) (?p Object) (?g Object) (?e Object) 
     (implies
     	 (not (= ?o ?p))
         (= (select2 (store2 ?h ?o ?f ?e) ?p ?g)
            (select2 ?h ?p ?g))))

  (forall (?h Heap) (?o Object) (?f Object) (?p Object) (?g Object) (?e Object) 
     (implies
     	 (not (= ?f ?g))
         (= (select2 (store2 ?h ?o ?f ?e) ?p ?g)
            (select2 ?h ?p ?g))))

  (forall (?o Object)
    (= ?o (int_to_object (object_to_int ?o))))

  (forall (?i Int)
    (= ?i (object_to_int (int_to_object ?i))))

  (forall (?o Object)
    (= ?o (array_to_object (object_to_array ?o))))

  (forall (?i Array)
    (= ?i (object_to_array (array_to_object ?i))))

  (forall (?a Array) (?i Object) (?e Object)
     (= (select (store ?a ?i ?e) ?i) ?e))

  (forall (?a Array) (?i Object) (?j Object) (?e Object) 
     (or (= ?i ?j) 
         (= (select (store ?a ?i ?e) ?j)
            (select ?a ?j))))

  ; (subtypes ...) is a partial order

  (forall (?t Object) (subtypes ?t ?t))

  (forall (?t Object) (?u Object) (?v Object)
     (implies (and (subtypes ?t ?u) (subtypes ?u ?v)) (subtypes ?t ?v)))

  (forall (?t Object) (?u Object)
     (implies (and (subtypes ?t ?u) (subtypes ?u ?t)) (= ?t ?u)))

  ; the integer part is not defined here
 )

 :notes 
 "
 "    
)
